import pygame
from locals import *
import util
import Events
import sound

class Character(pygame.sprite.Sprite):
    def __init__(self, image, start, world=None, coords=None):
        pygame.sprite.Sprite.__init__(self)
        self.start = start[0], start[1]
        self.world = world
        self.dir = U     
        self.health = 2
        self.max_health = 2

    def init_position(self):
        self.world_pos = [int(self.start[0]), int(self.start[1])]
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        self.p_world_pos = [self.start[0], self.start[1]]
        self.move_to = [self.start[0], self.start[1]]
        self.rect.left = self.screen_pos[0]
        self.rect.top = self.screen_pos[1]

    def init_images(self, image, coords=None):
        if coords is None:
            # default multi-sprite .bmp
            coords = ((1, 1, 19, 19),(21, 1, 19, 19),(41, 1, 19, 19),
                (1, 21, 19, 19),(41, 21, 19, 19),(1, 41, 19, 19),
                (21, 41, 19, 19),(41, 41, 19, 19))

        # eight direction subsurfaces: u,d,r,l,ur,ul,dr,dl
        self.full_image = util.get_image(image)
        self.full_image.set_colorkey((0, 138, 118))

        self.images = {
            UL: self.full_image.subsurface(coords[0]),
            U : self.full_image.subsurface(coords[1]),
            UR: self.full_image.subsurface(coords[2]),
            L : self.full_image.subsurface(coords[3]),
            R : self.full_image.subsurface(coords[4]),
            DL: self.full_image.subsurface(coords[5]),
            D : self.full_image.subsurface(coords[6]),
            DR: self.full_image.subsurface(coords[7])
        }
        
        self.image = self.images[self.dir]
        self.rect = pygame.Rect(self.image.get_rect())
        
    def update(self):
        if self.world.is_onscreen(self.world_pos, self.rect):
            self.image = self.images[self.dir]
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]
    
    def move(self):
        # Smooth movement
        dx = self.move_to[0] - self.world_pos[0]
        if abs(dx) > 1:
            self.world_pos[0] += dx * EASING
        else:
            self.world_pos[0] = self.move_to[0]
        dy = self.move_to[1] - self.world_pos[1]
        if abs(dy) > 1:
            self.world_pos[1] += dy * EASING
        else:
            self.world_pos[1] = self.move_to[1]

    def injure(self, dmg=1):
        self.health -= dmg
        if self.health <= 0:
            self.kill()
            return False
        else:
            return True

class Enemy(Character):
    def __init__(self, image, start, world=None):
        Character.__init__(self, image, start, world)
        coords = ((1,1,39,39),(41,1,39,39),(81,1,39,39),(1,41,39,39),
            (81,41,39,39),(1,81,39,39),(41,81,39,39),(81,81,39,39))
        self.init_images(image, coords)
        self.init_position()
        
        # choose a target offset from the main character.
        # each enemy chooses a random offset to prevent total clumping
        self.offset = random.randint(-200, 200), random.randint(-200, 200)
        self.target = [self.world.main_char.world_pos[0] + self.offset[0],
                       self.world.main_char.world_pos[1] + self.offset[1]]

        # world_pos center correction
        self.cx = self.rect.width / 2
        self.cy = self.rect.height / 2        
        self.center = self.world_pos[0] + self.cx, self.world_pos[1] + self.cy
        self.turret_image = util.get_image("turret")
        self.turret_image.set_colorkey((0,138,118))
        self.turret = self.turret_image
        self.draw_turret()
    
        self.count = 0
        self.max_count = random.randint(40, 60)
        self.health = 50
        self.max_health = 50
        self.bar = self.health / float(self.max_health) * self.rect.width
        self.v = .5
        self.me = "enemy"
        
        self.is_alive = True
        
    def update(self):
        self.count += 1
        pdir = self.dir
        self.dir = 0
        
        # recalc target
        self.target = [self.world.main_char.world_pos[0] + self.offset[0],
                       self.world.main_char.world_pos[1] + self.offset[1]]
        
        if self.target[0] < 0: self.target[0] = 0
        elif self.target[0] > WORLD_WIDTH: self.target[0] = WORLD_WIDTH
        if self.target[1] < 0: self.target[1] = 0
        elif self.target[1] > WORLD_HEIGHT: self.target[1] = WORLD_HEIGHT
        
        if self.target[0] < self.world_pos[0]:
            self.dir |= L
            self.move_to[0] -= self.v
        elif self.target[0] > self.world_pos[0]:
            self.dir |= R
            self.move_to[0] += self.v
        if self.target[1] < self.world_pos[1]:
            self.dir |= U
            self.move_to[1] -= self.v
        elif self.target[1] > self.world_pos[1]:
            self.dir |= D
            self.move_to[1] += self.v                
        
        if self.dir == 0:
            self.dir = pdir
        
        self.move()   
        
        self.angle = math.atan2(self.world_pos[0] - self.world.main_char.world_pos[0],
            self.world_pos[1] - self.world.main_char.world_pos[1])
        self.center = self.world_pos[0] + self.cx, self.world_pos[1] + self.cy
        
        self.rect = self.image.get_rect()
        
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        self.rect.left = self.screen_pos[0]
        self.rect.top = self.screen_pos[1]
        
        if self.world.is_onscreen(self.world_pos, self.rect):
            self.turret = pygame.transform.rotate(self.turret_image, self.angle * 57.2957795)
            self.draw_turret()
        
        if self.count >= self.max_count:
            self.count = 0
            self.world.fire_rocket(self, self.angle)

        #health
        pygame.draw.line(self.image, (138,0,0), (0,1), (self.bar, 1), 2)            

    def draw_turret(self):
        self.image = self.images[self.dir].copy()
        tr = self.turret.get_rect()
        self.image.blit(self.turret, 
            ((self.rect.width - tr.width) / 2.0,  (self.rect.height - tr.height) / 2.0))

    def injure(self, dmg=1):
        self.health -= dmg
        if self.health <= 0:
            self.kill()
            self.world.add_points(1000, self.world_pos)
            self.world.add_explosion(self.world_pos)
            sound.play_sound("enemy_explode")
            Events.trigger(self.me + "-die")  
            self.is_alive = False
            return False
        else:
            self.bar = self.health / float(self.max_health) * self.rect.width
            return True        

class NPC(Character):
    def __init__(self, image, start, world=None):
        Character.__init__(self, image, start, world)
        self.init_images(image)
        self.init_position()
        self.state = WANDER
        self.target = [start[0], start[1]]
        self.bad_guy = None
        self.v = 2

        self.sight = 75
        self.splode = 7
        self.convert = 0
        self.convert_rate = 4
        self.me = "character"
        
    def update(self):
        if self.state == HOLD:
            self.move()
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]
            return
        
        p_dir = self.dir
        v = self.v
        self.dir = 0
            
        if self.state == FOLLOW:
            me = self.world_pos
            mc = self.world.main_char.world_pos 
            if abs(me[0] - mc[0]) < self.sight and abs(me[1] - mc[1]) < self.sight:
                self.world.connections.append((self.rect.center, self.world.main_char.rect.center, 1))
                self.convert += self.convert_rate
                if self.convert > CONVERSION:
                    sound.play_sound("convert")
                    self.world.add_conversion(self.world_pos)
                    self.convert = 0
                    if self.dir == 0:
                        self.dir = p_dir
                    self.init_images('TankGreen')
                    self.state = ATTACK
                    e = [(dist(self.world_pos, e.world_pos), e) for e in self.world.enemies]
                    self.bad_guy = len(e) > 0 and min(e)[1] or None
            else:
                self.state = WANDER
        elif self.state == WANDER:
            if self.target[0] == self.world_pos[0] and \
               self.target[1] == self.world_pos[1]:
                
                # choose new wander destination, ideally on the same screen as 
                # the main character
                self.target[0] = random.randint(
                    int(self.world.main_char.world_pos[0]-HALF_SCREEN_WIDTH),
                    int(self.world.main_char.world_pos[0]+HALF_SCREEN_WIDTH))
                self.target[1] = random.randint(
                    int(self.world.main_char.world_pos[1]-HALF_SCREEN_HEIGHT),
                    int(self.world.main_char.world_pos[1]+HALF_SCREEN_HEIGHT))
                    
                if self.target[0] < 0: self.target[0] = 0
                elif self.target[0] > WORLD_WIDTH: self.target[0] = WORLD_WIDTH
                if self.target[1] < 0: self.target[1] = 0
                elif self.target[1] > WORLD_HEIGHT: self.target[1] = WORLD_HEIGHT
        
            if self.target[0] < self.move_to[0]:
                self.dir |= L
            elif self.target[0] > self.move_to[0]:
                self.dir |= R
            if self.target[1] < self.move_to[1]:
                self.dir |= U
            elif self.target[1] > self.move_to[1]:
                self.dir |= D
    
            # scan for main_char
            me = self.world_pos
            mc = self.world.main_char.world_pos
            if abs(me[0] - mc[0]) < self.sight and abs(me[1] - mc[1]) < self.sight:
                self.state = FOLLOW
                self.world.connections.append((self.rect.center, self.world.main_char.rect.center, 1))
            else:
                self.state = WANDER
        elif self.state == ATTACK and self.bad_guy:
            if not self.bad_guy.is_alive:
                e = [(dist(self.world_pos, e.world_pos), e) for e in self.world.enemies]
                self.bad_guy = len(e) > 0 and min(e)[1] or None
            if not self.bad_guy:
                return
            me = self.world_pos
            mc = self.bad_guy.center
            if abs(me[0] - mc[0]) < self.splode and abs(me[1] - mc[1]) < self.splode:
                self.kill()
                self.world.add_small_explosion(self.world_pos)
                self.bad_guy.injure(4)
                self.world.add_points(500, self.world_pos)
                Events.trigger(self.me + "-die")                
                sound.play_sound("hit_enemy", .4)
                return
            
            self.target = self.bad_guy.center           
            
            if self.target[0] < self.move_to[0]:
                self.dir |= L
            elif self.target[0] > self.move_to[0]:
                self.dir |= R
            if self.target[1] < self.move_to[1]:
                self.dir |= U
            elif self.target[1] > self.move_to[1]:
                self.dir |= D
        else:
            e = [(dist(self.world_pos, e.world_pos), e) for e in self.world.enemies]
            self.bad_guy = len(e) > 0 and min(e)[1] or None
            
        if self.dir == 0:
            self.dir = p_dir
        else: 
            if self.dir & L:
                if self.move_to[0] - self.target[0] > v:
                    self.move_to[0] -= v
                else: 
                    self.move_to[0] = self.target[0]
            elif self.dir & R:            
                if self.target[0] - self.move_to[0] > v:
                    self.move_to[0] += v
                else: 
                    self.move_to[0] = self.target[0]
            if self.dir & U:    
                if self.move_to[1] - self.target[1] > v:
                    self.move_to[1] -= v
                else: 
                    self.move_to[1] = self.target[1]                        
            elif self.dir & D:
                if self.target[1] - self.move_to[1] > v:
                    self.move_to[1] += v
                else: 
                    self.move_to[1] = self.target[1]
                       
        self.move()

        self.image = self.images[self.dir]
        if self.convert > 0:
            pygame.draw.line(self.image,
                             (28,195,0),
                             (0,17),
                             (float(self.convert) / CONVERSION * self.rect.width, 17),
                             2)
        self.screen_pos = self.world.world_to_screen(self.world_pos)
        self.rect.left = self.screen_pos[0]
        self.rect.top = self.screen_pos[1]

    
class Tank(Character):
    def __init__(self, start, world=None):
        Character.__init__(self, HERO_TANK, start, world)
        self.init_images(HERO_TANK)
        self.init_position()
        self.v = 5
        self.me = "hero"
        self.health = HP
        self.max_health = HP
                        
    def update(self):
        p_dir = self.dir
        self.dir = 0
        
        self.p_world_pos[0] = self.world_pos[0]
        self.p_world_pos[1] = self.world_pos[1]

        # Move destination
        if KEYS[K_LEFT] and self.move_to[0] > 0:
            self.move_to[0] -= self.v
        elif KEYS[K_RIGHT] and self.move_to[0] + self.rect.width < WORLD_WIDTH:
            self.move_to[0] += self.v

        if KEYS[K_UP] and self.move_to[1] > 0:
            self.move_to[1] -= self.v
        elif KEYS[K_DOWN] and self.move_to[1] + self.rect.height < WORLD_HEIGHT:
            self.move_to[1] += self.v
    
        # Choose image
        if self.move_to[0] < self.world_pos[0] and KEYS[K_LEFT]: 
            self.dir |= L
        elif self.move_to[0] > self.world_pos[0] and KEYS[K_RIGHT]:
            self.dir |= R
            
        if self.move_to[1] > self.world_pos[1] and KEYS[K_DOWN]:
            self.dir |= D
        elif self.move_to[1] < self.world_pos[1] and KEYS[K_UP]:
            self.dir |= U            

        if self.dir > 0:
            self.image = self.images[self.dir]
        else:
            self.dir = p_dir

        if self.move_to[0] != self.world_pos[0] or\
           self.move_to[1] != self.world_pos[1]:
            self.move()
            self.screen_pos = self.world.world_to_screen(self.world_pos)
            self.rect.left = self.screen_pos[0]
            self.rect.top = self.screen_pos[1]   
        

