"""Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "lib"
directory.
"""

import pygame
import sys

# splush packages (with modifications)
import data
import util
from menu import Menu
from locals import *

#gather Classes
from Character import NPC, Enemy
from Events import on_key_press, on_key_release, event_handlers
from World import World, start_level
import Animation
import sound

try:
    import psyco
    psyco.full()
except ImportError:
    print "no dice, shoulda' installed psyco."
    pass

def main():
    # These should always be run first for the other included modules to work
    util.init()
    pygame.init()

    screen = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT)) #, FULLSCREEN | HWSURFACE | DOUBLEBUF)
    pygame.display.set_caption('gather')
    pygame.mouse.set_visible = False
    clock = pygame.time.Clock()
    Animation.init_anims()

    def level_up(surf=None):
        """Add new level definitions here to extend normal gameplay. 
        
        for example:
            Menu(screen, ['continue'], surf, '"explod-o-rama"', "level_start").run()
            yield start_level(world, 90, 20)
        """
        bk = pygame.Surface((SCREEN_WIDTH, SCREEN_HEIGHT))
        bk.fill((0,0,0))
        bk.set_alpha(0)
        clock = pygame.time.Clock()
        
        Menu(screen, ['continue'], None, '"easy"', "level_start").run()
        yield start_level(world, 15, 1)
        sound.play_sound("level_start")
        
        # fade out
        aph = 0
        while aph < 180:
            screen.blit(bk, (0,0))
            pygame.display.flip()
            clock.tick(FPS)
            aph += 10
            bk.set_alpha(aph)
        
        Menu(screen, ['continue'], screen, '"little tougher"', "level_start").run()
        yield start_level(world, 20, 4)
        sound.play_sound("level_start")

        # fade out
        aph = 0
        while aph < 180:
            screen.blit(bk, (0,0))
            pygame.display.flip()
            clock.tick(FPS)
            aph += 10
            bk.set_alpha(aph)

        Menu(screen, ['continue'], screen, '"lotsa rockets"', "level_start").run()
        yield start_level(world, 35, 8)
        sound.play_sound("level_start")
        
        # fade out
        aph = 0
        while aph < 180:
            screen.blit(bk, (0,0))
            pygame.display.flip()
            clock.tick(FPS)
            aph += 10
            bk.set_alpha(aph)
        
        Menu(screen, ['continue'], surf, '"mayhem"', "level_start").run()
        yield start_level(world, 65, 12)
    
    world = World(screen)
    
    def add_c():
        while len(world.characters) < world.char_max:
            world.add_character(NPC('TankRed', 
                # random screen's edge respawn
                random.randint(0,1) == 1 and \
                    # right edge 
                    (WORLD_WIDTH + 20, random.randint(0, WORLD_HEIGHT)) or \
                    # bottom edge
                    (random.randint(0, WORLD_WIDTH), WORLD_HEIGHT + 20)     
                ,world))
        if world.game_mode == GAME_SURVIVAL:
            # increasing NPC count
            if world.char_max < len(world.enemies) * 8:
                world.char_max += 10
        elif world.game_mode == GAME_NORMAL:
            # decreasing NPC count
            if world.char_max > 20:
                world.char_max = len(self.enemies) * 8
        # ignore NPC count in GAME_CUSTOM
    def add_e():
        world.add_enemy(Enemy('Transport', 
            (random.randint(0, WORLD_WIDTH), random.randint(0, WORLD_HEIGHT)),world))
        world.add_enemy(Enemy('Transport', 
            (random.randint(0, WORLD_WIDTH), random.randint(0, WORLD_HEIGHT)),world))

    def show_menu(surf=None):
        "in game menu"
        sound.play_music("menu_loop")
        vals = Menu(screen, ['return to game', 'main menu', 'exit'], surf, 'gather').run()
        if vals['exit']:
            return QUITTER
        elif vals['main menu']:
            return MAIN_MENU
        else: # return to game
            sound.play_music("game_song_02")

    def end_menu():
        "this event is triggered from inside the world.update() function"
        sound.play_music("menu_loop")
        
        bk = pygame.Surface((SCREEN_WIDTH, SCREEN_HEIGHT))
        bk.fill((0,0,0))
        bk.set_alpha(0)
        clock = pygame.time.Clock()
        
        # fade out
        aph = 0
        while aph < 255:
            screen.blit(bk, (0,0))
            pygame.display.flip()
            clock.tick(FPS)
            aph += 10
            bk.set_alpha(aph)
        
        vals = Menu(screen, ["Final Score: %i" % world.score], screen.copy(), "thanks for playing").run()
        return main_menu()        
            
    def hero_die():
        world.lives -= 1
        world.main_char.kill()
        world.main_char_group.empty()
        del world.main_char
        if world.lives <= 0:
            return end_menu()
        else:
            world.respawn()
            return False

    event_handlers[K_SPACE] = lambda : world.fire_ball(world.main_char)
    event_handlers[K_m] = lambda : show_menu(world.screen.copy())
    event_handlers["character-die"] = add_c
    event_handlers["hero-die"] = hero_die
    event_handlers["end-game"] = end_menu
    
    def main_menu():
        sound.play_music("menu_loop")
        while 1:
            m = Menu(screen, ['start game', 'custom game', 'survival mode', 'exit'], None, 'gather')
            vals = m.run()
            _quit = vals['exit']
            event_handlers["enemy-die"] = lambda : ""
            if vals['start game']:
                world.score = 0
                world.render_score()
                lu = level_up()
                world.game_mode = GAME_NORMAL
                event_handlers["level-up"] = lu.next
                lu.next()
                break
            elif vals['custom game']:
                # custom game
                vals = Menu(screen, 
                    [('enemies', 5), ('friends', 20), ('lives', 3), 'start game', 'return'], 
                    None, 'custom game setup').run()
                if vals['return'] or 'exit' in vals and vals['exit']:
                    continue
                else:
                    world.score = 0
                    world.render_score()
                    world.lives = vals['lives']
                    world.game_mode = GAME_CUSTOM
                    Menu(screen, ['continue'], 
                        None, '"let the explosions begin"', "level_start").run()
                    event_handlers["level-up"] = end_menu
                    start_level(world, vals['friends'], vals['enemies'])
                    break
            elif vals['survival mode']:
                world.score = 0
                world.render_score()
                world.game_mode = GAME_SURVIVAL
                event_handlers["level-up"] = end_menu
                event_handlers["enemy-die"] = add_e
                start_level(world, 1, 1)
                break
            elif _quit:
                return QUITTER
        return False
        
    _quit = main_menu() 
    show_main = False
    while not _quit:
        if show_main:
            retval = main_menu()
            if retval == QUITTER:
                _quit = True                
            show_main = False
    
        clock.tick(FPS)
        for e in pygame.event.get():
            if e.type is QUIT: 
                _quit = True
            elif e.type is KEYDOWN and e.key == K_ESCAPE: 
                # trigger the in-game menu with ESC only if the game is currently running.
                retval = event_handlers[K_m]()
                if retval == QUITTER:
                    _quit = True
                elif retval == MAIN_MENU:
                    show_main = True
            elif e.type is KEYDOWN:
                retval = on_key_press(e.key, e.mod)
                if retval == QUITTER:
                    _quit = True
                elif retval == MAIN_MENU:
                    show_main = True
            elif e.type is KEYUP:
                on_key_release(e.key, e.mod)

        if not _quit:
            retval = world.update()
            if retval == MAIN_MENU: #should only trigger on endgame
                show_main = True
            elif retval == QUITTER:
                _quit = True
            else:
                world.draw_all()
        
        pygame.display.flip()

    return

