""" This module contains generic utility functions.
The comments give a basic idea of each function's purpose,
and they can also be found from the WWW documentation."""

import pygame
#from pygame.locals import *
from string import lower
import os
import codecs
import datetime

from locals import *

import data

from variables import Variables

class Util:
    """A storage class for cached images, variables used by the
    utility functions and such."""
    pygame.font.init()
    smallfont = pygame.font.Font(data.filepath(
        os.path.join("fonts", "datat.ttf")), FONT_SIZE_SMALL)
    bigfont = pygame.font.Font(data.filepath(
        os.path.join("fonts", "chintzy.ttf")), FONT_SIZE_BIG)
    cached_text_images = {}
    fade_state = FADE_STATE_BLACK
    blackscreen = pygame.Surface((SCREEN_WIDTH, SCREEN_HEIGHT))
    image_path = data.filepath(os.path.join('images', '%s.%s'))

def get_image(filename, ext='bmp'):
    return pygame.image.load(Util.image_path % (filename, ext)).convert()
    
def init():
    Variables.vdict["sound"] = True
    Variables.vdict["fullscreen"] = True
    
def render_text(string, color=COLOR_GUI, bgcolor=COLOR_GUI_BG, 
   size=FONT_SIZE_SMALL):
    """Render fancy-looking text. The function also handles caching of text
    images. Returns a pygame surface containing the rendered text.
    The text is rendered with slight edges to make it look more readable on a
    colorful background.

    The constant colors can be found from locals.py."""
    if Util.cached_text_images.has_key(string + str(color) + str(bgcolor)):
        final_image = Util.cached_text_images[string + str(color) +
            str(bgcolor)]
    else:
        if size == FONT_SIZE_SMALL:
            text_image_bg = Util.smallfont.render(string, True, bgcolor)
            text_image_fg = Util.smallfont.render(string, True, color)
        elif size == FONT_SIZE_BIG:
            text_image_bg = Util.bigfont.render(string, True, bgcolor)
            text_image_fg = Util.bigfont.render(string, True, color)
        rect = text_image_bg.get_rect()
        final_image = pygame.Surface(
            (rect.width + 2, rect.height + 2)).convert_alpha()

        final_image.fill((0,0,0,0))

        final_image.blit(text_image_bg, rect)
        final_image.blit(text_image_bg, (1,0))
        final_image.blit(text_image_bg, (2,0))

        final_image.blit(text_image_bg, (0,2))
        final_image.blit(text_image_bg, (1,2))
        final_image.blit(text_image_bg, (2,2))

        final_image.blit(text_image_fg, (1,1))
        Util.cached_text_images[string + str(color) + str(bgcolor)] = final_image
    return final_image

def render_score(scr):
    return render_text("%08i" % int(scr), size=FONT_SIZE_SMALL)

def fade_to_black(screen, fade_target):
    """Fades the screen to black. Both fade-in and fade-out.
    Returns true if the fading has finished.
    fade_target should be an integer (0-255)
    * 255 = FADE_STATE_BLACK : The display is all black
    * 0 = FADE_STATE_NONE : The display is not faded at all"""
    if Util.fade_state > fade_target:
        Util.fade_state += int(255 / (FPS * FADE_IN))
        if Util.fade_state < fade_target:
            Util.fade_state = fade_target
    if Util.fade_state < fade_target:
        Util.fade_state -= -int(255 / (FPS * FADE_OUT))
        if Util.fade_state > fade_target:
            Util.fade_state = fade_target
    if Util.fade_state > FADE_STATE_NONE:
        Util.blackscreen.set_alpha(Util.fade_state)
        screen.blit(Util.blackscreen, screen.get_rect())
    return (Util.fade_state == fade_target)

def ease(target, current):
    return current + ((target - current) * EASING)
